import 'dart:async';
import 'dart:math';
// import 'dart:html';

import 'package:dart_arango_min/dart_arango_min.dart';
import 'package:test/test.dart';
import "dart:io";
import 'test_conf.dart';
import "dart:convert";

void main() {
  group('Client can:', () {
    final sch = dbscheme;
    final h = dbhost;
    final p = dbport;
    const system_db = '_system';
    const test_db = 'test_temp_db';
    const test_collection = 'test_temp_collection';
    String test_document_key;
    String test_document_rev;
    final u = dbuser;
    final ps = dbpass;
    const realm = "";

    ArangoDBClient client_system_db = ArangoDBClient(
        scheme: sch,
        host: h,
        port: p,
        db: system_db,
        user: u,
        pass: ps,
        realm: realm);

    // setUp(() {
    //   client = ArangoDBClient(
    //       scheme: sch,
    //       host: h,
    //       port: p,
    //       db: system_db,
    //       user: u,
    //       pass: ps,
    //       realm: realm);
    // });

    test('Get current db info', () async {
      var answer = await client_system_db.currentDatabase();
      expect(answer, contains('error'));
      if (answer['error'] == true) {
        print(answer);
      }
      expect(answer, contains('result'));
      expect(answer['result'] as Map, contains('name'));
      expect(answer['result']['name'] as String, equals(system_db));
      expect(answer['result'] as Map, contains('isSystem'));
      expect(answer['result']['isSystem'] as bool, equals(true));
    });

    test('List of accessible databases', () async {
      var answer = await client_system_db.userDatabases();
      if (answer['error'] == true) {
        print(answer);
      }
      expect(answer['error'] as bool, equals(false));
      expect(answer, contains('result'));
      expect(answer['result'] as List, contains(system_db));
    });

    test('List of all existing databases', () async {
      var answer = await client_system_db.existingDatabases();
      if (answer['error'] == true) {
        print(answer);
      }
      expect(answer['error'] as bool, equals(false));
      expect(answer, contains('result'));
      expect(answer['result'] as List, contains(system_db));
    });

    test('Create a database', () async {
      // first, ask about all databases
      var databases = await client_system_db.existingDatabases();
      // skip test if test database already exists
      if (databases.containsKey('result') &&
          (databases['result'] as List).contains(test_db)) {
        print(
            "Skip test for creating database because database ${test_db} already exists.");
        return;
      }
      // creating the test database
      var answer = await client_system_db.createDatabase({
        "name": test_db,
        "users": [
          {
            "username": u,
            "passwd": ps,
          },
        ],
      });
      expect(answer, contains('error'));
      expect(answer['error'] as bool, equals(false));
      expect(answer, contains('result'));
      expect(answer['result'], equals(true));
    });

    // changing current database
    ArangoDBClient test_db_client = ArangoDBClient(
        scheme: sch,
        host: h,
        port: p,
        db: test_db,
        user: u,
        pass: ps,
        realm: realm);

    test('create collection', () async {
      var all_collections_answ = await test_db_client.allCollections();
      var already_exists = (all_collections_answ['result'] as List)
          .any((coll) => coll['name'] == test_collection);
      if (already_exists) {
        print(
            'Skip creating collection ${test_collection} bexause it is exists');
      } else {
        var answer =
            await test_db_client.createCollection({"name": test_collection});
        if (answer['error'] == true) {
          print(answer);
        }
        expect(answer['error'], false);
        expect(answer['name'] as String, test_collection);
      }
    });

    test('truncate collection', () async {
      var answer = await test_db_client.truncateCollection(test_collection);
      if (answer['error'] == true) {
        print(answer);
      }
      expect(answer['error'], false);
      expect(answer['name'] as String, test_collection);
    });

    test('get collection info', () async {
      var answer = await test_db_client.collectionInfo(test_collection);
      if (answer['error'] == true) {
        print(answer);
      }
      expect(answer['error'], false);
      expect(answer['name'] as String, test_collection);
      expect(answer, contains('status'));
    });

    test('get collection properties', () async {
      var answer = await test_db_client.collectionProperties(test_collection);
      if (answer['error'] == true) {
        print(answer);
      }
      expect(answer['error'], false);
      expect(answer, contains('waitForSync'));
      expect(answer, contains('keyOptions'));
    });

    test('get count of documents in collection', () async {
      var answer = await test_db_client.documentsCount(test_collection);
      if (answer['error'] == true) {
        print(answer);
      }
      expect(answer['error'], false);
      expect(answer, contains('count'));
    });

    test('get statistics for a collection', () async {
      var answer = await test_db_client.collectionStatistics(test_collection);
      if (answer['error'] == true) {
        print(answer);
      }
      expect(answer['error'], false);
      expect(answer, contains('figures'));
    });

    test('get collection revision id', () async {
      var answer = await test_db_client.collectionRevisionId(test_collection);
      if (answer['error'] == true) {
        print(answer);
      }
      expect(answer['error'], false);
      expect(answer, contains('revision'));
    });

    test('get collection checksum', () async {
      var answer = await test_db_client.collectionChecksum(test_collection);
      if (answer['error'] == true) {
        print(answer);
      }
      expect(answer['error'], false);
      expect(answer, contains('checksum'));
    });

    test('get all collections', () async {
      var answer = await test_db_client.allCollections();
      if (answer['error'] == true) {
        print(answer);
      }
      expect(answer['error'], false);
      expect(answer, contains('result'));
    });

    test('create document', () async {
      var answer = await test_db_client
          .createDocument(test_collection, {"Hello": "World"});
      if (answer['error'] == true) {
        print(answer);
      }
      expect(answer,
          allOf(contains('_id'), contains('_key'), contains('_rev')));
      // save document key for next test:
      test_document_key = answer['_key'] as String;
    });
    test('getDocumentByKey returns Map with _key', () async {
      var answer = await test_db_client.getDocumentByKey(
          test_collection, test_document_key);
      if (answer['error'] == true) {
        print(answer);
      }
      expect(answer, contains("_key"));
    });

    test('getDocumentByKey can require document revision', () async {
      // first get the doc:
      var answer = await test_db_client.getDocumentByKey(
          test_collection, test_document_key);
      // get its revision:
      expect(answer, contains('_rev'));
      if (answer['error'] == true) {
        print(answer);
      }
      // save its revision:
      test_document_rev = answer['_rev'] as String;

      // now try to get the document with not matched revision:
      var empty_answer = await test_db_client.getDocumentByKey(
          test_collection, test_document_key,
          if_none_match_revision: test_document_rev);

      // because last revision equals ${test_document_rev} server answer will empty:
      expect(empty_answer, equals({}));

      // get document only required revision:
      var answer_with_revision = await test_db_client.getDocumentByKey(
          test_collection, test_document_key,
          if_match_revision: test_document_rev);

      expect(answer_with_revision, contains('_rev'));
      if (answer_with_revision['error'] == true) {
        print(answer_with_revision);
      }
      expect(answer_with_revision['_rev'], equals(test_document_rev));

      // try get doc with noexists revision:
      var not_exist_answer = await test_db_client.getDocumentByKey(
          test_collection, test_document_key,
          if_match_revision: "my_wrong_rev");

      // we will get error responce:
      expect(not_exist_answer, contains('error'));
      expect(not_exist_answer['error'], equals(true));
      expect(not_exist_answer, contains('code'));
      expect(not_exist_answer['code'], equals(412));
    });

    test('update document', () async {
      // update document with returnNew:
      var update_answer = await test_db_client.updateDocument(
          test_collection, test_document_key, {"Hello": "University"},
          queryParams: {"returnNew": "true"});

      if (update_answer['new']['Hello'] != 'University') print(update_answer);

      expect(update_answer["new"]["Hello"], "University");

      // save revision:
      test_document_rev = update_answer['_rev'] as String;

      var matched_update_answer = await test_db_client.updateDocument(
          test_collection, test_document_key, {"Hello": "Underground"},
          if_match_revision: test_document_rev);

      // document was updated:
      expect(matched_update_answer, contains('_oldRev'));
      expect(matched_update_answer['_oldRev'], equals(test_document_rev));

      // try to update not matched revision:
      var not_matched_update_answer = await test_db_client.updateDocument(
          test_collection,
          test_document_key,
          {"Bad trying": "because bad revision"},
          if_match_revision: "my_bad_rev");

      // we will get error in answer:
      expect(not_matched_update_answer, contains('error'));
      expect(not_matched_update_answer['error'], equals(true));
      expect(not_matched_update_answer['code'], equals(412));
    });

    test('create multiple documents', () async {
      var answer = await test_db_client.createDocument(test_collection, [
        {"Hello": "Earth"},
        {"Hello": "Venus"}
      ]);
      // 2 documents was inserted:
      expect((answer as List).length, equals(2));
      // about first document:
      expect(answer[0],
          allOf(contains('_id'), contains('_key'), contains('_rev')));
    });

    // td: Create a document in a collection named products with a collection-level waitForSync value of false.
    // td: Create a document in a collection with a collection-level waitForSync value of false, but using the waitForSync query parameter
    // td: test with illegal document ...
    // td:  test creating document with returnNew parameter...
    // td: create doc with the same _key and so get new revision ...

    test('drop collection', () async {
      var answer = await test_db_client.dropCollection(test_collection);
      if (answer['error'] == true) {
        print(answer);
      }
      expect(answer['error'] as bool, false);
    });

    // back to _system db
    test('drop test database', () async {
      var answer = await client_system_db.dropDatabase(test_db);
      if (answer['error'] == true) {
        print(answer);
      }
      expect(answer, contains('result'));
      expect(answer['result'], equals(true));
    });

// td: определиться какой клиент использовать

    //   test('query_first_batch() returns Map with "result", "error", "hasMore".',
    //       () async {
    //     var data = {"query": "for d in aaa return d"};
    //     var j = await client
    //         .query_first_batch(data); // потом же изменим запрос, правда?
    //     // print(j);
    //     expect(j, contains("result"));
    //     expect(j, contains("error"));
    //     expect(j, contains("hasMore"));
    //     expect(j["error"], false);
    //   });

    //   test('first batch of multiple-batched cursor has "hasMore"==true and "id"',
    //       () async {
    //     var data = {"query": "for d in aaa return d", "batchSize": 1};
    //     var j = await client.query_first_batch(data);
    //     expect(j, contains("hasMore"));
    //     expect(j["hasMore"], equals(true));
    //     expect(j, contains("id"));
    //   });

// test('query_first_batch() returns Map with "result", "error", "hasMore".',
    //     () async {
    //   var data = {"query": "for d in aaa return d"};
    //   var j = await client
    //       .query_first_batch(data); // потом же изменим запрос, правда?
    //   // print(j);
    //   expect(j, contains("result"));
    //   expect(j, contains("error"));
    //   expect(j, contains("hasMore"));
    //   expect(j["error"], false);
    // });

    // test('batch of multiple-batched cursor has "hasMore"==true|false and "id"',
    //     () async {
    //   var data = {"query": "for d in aaa return d", "batchSize": 1};
    //   var j = await client.query_first_batch(data);
    //   expect(j, contains("hasMore"));
    //   expect(j["hasMore"], equals(true));
    //   expect(j, contains("id"));

    //   // fetching all next batches
    //   var j2;
    //   do {
    //     var cursorId = int.parse(j['id'] as String);
    //     j2 = await client.fetch_next_batch(cursorId);
    //   } while (j2['hasMore'] as bool == true);

    //   //  now j2 contains last batch
    //   expect(j2['hasMore'], equals(false));
    // });

    // test('query', () async {
    //   var data = {"query": "for d in aaa return d", "batchSize": 1};
    //   var batches = await client.query(data);
    //   await for (var batch in batches) {
    //     List records = batch['result'] as List;
    //     for (Map document in records){
    //       expect(document, contains('_key'));
    //     }
    //   }
    // });
  });
}
